@@ -698,7 +698,7 @@ grn_geo_table_sort(grn_ctx *ctx, grn_obj *table, int offset, int limit,
     unsigned int size;
     grn_rc rc;
     size = grn_table_size(ctx, table);
-    rc = grn_normalize_offset_and_limit(ctx, size, &offset, &limit);
+    rc = grn_output_range_normalize(ctx, size, &offset, &limit);
     if (rc != GRN_SUCCESS) {
       ERR(rc,
           "[sort][geo] failed to normalize offset and limit: "
@@ -2015,7 +2015,9 @@ grn_geo_select_in_rectangle_callback(grn_ctx *ctx, grn_posting *posting,
                                      void *user_data)
 {
   grn_geo_select_in_rectangle_data *data = user_data;
-  grn_ii_posting_add(ctx, posting, data->res, data->op);
+  grn_posting add_posting = *posting;
+  add_posting.weight += 1;
+  grn_ii_posting_add(ctx, &add_posting, data->res, data->op);
   return GRN_TRUE;
 }
 
