@@ -26,7 +26,6 @@
 
 static uint64_t grn_index_sparsity = 10;
 static grn_bool grn_index_chunk_split_enable = GRN_TRUE;
-static size_t grn_index_column_diff_cache_size_max = 256;
 
 void
 grn_index_column_init_from_env(void)
@@ -57,21 +56,6 @@ grn_index_column_init_from_env(void)
       grn_index_chunk_split_enable = GRN_TRUE;
     }
   }
-
-  {
-    char grn_index_column_diff_cache_size_max_env[GRN_ENV_BUFFER_SIZE];
-    grn_getenv("GRN_INDEX_COLUMN_DIFF_CACHE_SIZE_MAX",
-               grn_index_column_diff_cache_size_max_env,
-               GRN_ENV_BUFFER_SIZE);
-    if (grn_index_column_diff_cache_size_max_env[0]) {
-      uint64_t size;
-      errno = 0;
-      size = strtoull(grn_index_column_diff_cache_size_max_env, NULL, 0);
-      if (errno == 0 && size > 0) {
-        grn_index_column_diff_cache_size_max = size;
-      }
-    }
-  }
 }
 
 grn_inline static void
@@ -87,7 +71,8 @@ grn_index_column_build_call_hook(grn_ctx *ctx,
   if (hooks) {
     /* todo : grn_proc_ctx_open() */
     grn_obj id_, flags_;
-    grn_proc_ctx pctx = {{0}, hooks->proc, NULL, hooks, hooks, PROC_INIT, 4, 4};
+    grn_proc_ctx pctx;
+    grn_proc_ctx_init(&pctx, hooks, 4, 4);
     GRN_UINT32_INIT(&id_, 0);
     GRN_UINT32_INIT(&flags_, 0);
     GRN_UINT32_SET(ctx, &id_, id);
@@ -259,16 +244,44 @@ grn_index_column_rebuild(grn_ctx *ctx, grn_obj *index_column)
   GRN_API_RETURN(ctx->rc);
 }
 
-static const char *postings_column_name = "postings";
+typedef enum {
+  GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS,
+  GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS,
+  GRN_INDEX_COLUMN_DIFF_TYPE_LOOK_AHEAD,
+} grn_index_column_diff_type;
+/* Except LOOK_AHEAD */
+#define N_GRN_INDEX_COLUMN_DIFF_PERSISTENT_TYPES 2
+
+static const char *
+grn_index_column_diff_type_to_string(grn_index_column_diff_type type)
+{
+  switch (type) {
+  case GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS :
+    return "remains";
+  case GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS :
+    return "missings";
+  case GRN_INDEX_COLUMN_DIFF_TYPE_LOOK_AHEAD :
+    return "look-ahead";
+  default :
+    return "unknown";
+  }
+}
+
 static const char *remains_column_name = "remains";
 static const char *missings_column_name = "missings";
 
 typedef struct {
   grn_id token_id;
-  grn_id diff_id;
-  grn_obj postings;
-  size_t offset;
-} grn_index_column_diff_cache;
+  grn_obj remains;
+  grn_obj missings;
+  grn_obj look_ahead;
+  size_t look_ahead_offset;
+  grn_posting last_posting;
+  grn_ii_cursor *cursor;
+  bool need_cursor_next;
+  size_t n_postings[N_GRN_INDEX_COLUMN_DIFF_PERSISTENT_TYPES];
+  size_t thresholds[N_GRN_INDEX_COLUMN_DIFF_PERSISTENT_TYPES];
+} grn_index_column_diff_posting_list;
 
 typedef struct {
   grn_obj *lexicon;
@@ -284,7 +297,7 @@ typedef struct {
   grn_obj *source_table;
   grn_obj source_columns;
   grn_obj *diff;
-  grn_obj *postings;
+  grn_hash *posting_lists;
   grn_obj *remains;
   grn_obj *missings;
   struct {
@@ -306,8 +319,6 @@ typedef struct {
     grn_timeval start_time;
     grn_timeval previous_time;
   } progress;
-  grn_index_column_diff_cache *cache;
-  size_t cache_size;
 } grn_index_column_diff_data;
 
 static void
@@ -317,45 +328,6 @@ grn_index_column_diff_data_init(grn_ctx *ctx,
   GRN_PTR_INIT(&(data->source_columns), GRN_OBJ_VECTOR, GRN_ID_NIL);
   GRN_VOID_INIT(&(data->buffers.value));
   GRN_UINT32_INIT(&(data->buffers.postings), GRN_OBJ_VECTOR);
-
-  data->cache = NULL;
-  data->cache_size = 0;
-}
-
-static void
-grn_index_column_diff_data_init_cache(grn_ctx *ctx,
-                                      grn_index_column_diff_data *data)
-{
-  size_t cache_size;
-  const size_t n_tokens = grn_table_size(ctx, data->lexicon);
-  if (n_tokens == 0) {
-    cache_size = 1;
-  } else if (n_tokens < grn_index_column_diff_cache_size_max) {
-    cache_size = n_tokens;
-  } else {
-    cache_size = n_tokens * 0.01;
-    if (cache_size >= grn_index_column_diff_cache_size_max) {
-      cache_size = grn_index_column_diff_cache_size_max;
-    }
-  }
-
-  data->cache = GRN_MALLOCN(grn_index_column_diff_cache, cache_size);
-  if (!data->cache) {
-    char message[GRN_CTX_MSGSIZE];
-    grn_strcpy(message, GRN_CTX_MSGSIZE, ctx->errbuf);
-    ERR(GRN_NO_MEMORY_AVAILABLE,
-        "[index-column][diff] failed to allocate cache: <%" GRN_FMT_SIZE ">: %s",
-        cache_size,
-        message);
-    return;
-  }
-  for (size_t i = 0; i < cache_size; i++) {
-    data->cache[i].token_id = GRN_ID_NIL;
-    data->cache[i].diff_id = GRN_ID_NIL;
-    GRN_UINT32_INIT(&(data->cache[i].postings), GRN_OBJ_VECTOR);
-    data->cache[i].offset = 0;
-  }
-  data->cache_size = cache_size;
 }
 
 static void
@@ -375,11 +347,6 @@ grn_index_column_diff_data_fin(grn_ctx *ctx,
 
   GRN_OBJ_FIN(ctx, &(data->buffers.value));
   GRN_OBJ_FIN(ctx, &(data->buffers.postings));
-
-  for (size_t i = 0; i < data->cache_size; i++) {
-    GRN_OBJ_FIN(ctx, &(data->cache[i].postings));
-  }
-  GRN_FREE(data->cache);
 }
 
 static void
@@ -497,271 +464,511 @@ grn_index_column_diff_progress(grn_ctx *ctx,
   }
 }
 
-static size_t
-grn_index_column_diff_cache_compute_key(grn_ctx *ctx,
-                                        grn_index_column_diff_data *data,
-                                        grn_id token_id)
+typedef enum {
+  GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL,
+  GRN_INDEX_COLUMN_DIFF_COMPARED_LESS,
+  GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER,
+} grn_index_column_diff_compared;
+
+static char
+grn_index_column_diff_compared_to_char(grn_index_column_diff_compared compared)
 {
-  return token_id % data->cache_size;
-}
-
-static void
-grn_index_column_diff_cache_flush(grn_ctx *ctx,
-                                  grn_index_column_diff_data *data,
-                                  size_t cache_key)
-{
-  const grn_id cached_token_id = data->cache[cache_key].token_id;
-  if (cached_token_id == GRN_ID_NIL) {
-    return;
-  }
-
-  grn_obj *new_postings = &(data->buffers.postings);
-  GRN_BULK_REWIND(new_postings);
-  grn_obj *postings = &(data->cache[cache_key].postings);
-  const size_t n_posting_elements = data->n_posting_elements;
-  const size_t offset = data->cache[cache_key].offset;
-  const uint32_t *postings_raw =
-    ((uint32_t *)GRN_BULK_HEAD(postings)) +
-    (n_posting_elements * offset);
-  const size_t postings_raw_size =
-    sizeof(uint32_t) *
-    (GRN_UINT32_VECTOR_SIZE(postings) - (n_posting_elements * offset));
-  grn_bulk_write(ctx,
-                 new_postings,
-                 (const char *)postings_raw,
-                 postings_raw_size);
-  grn_obj_set_value(ctx,
-                    data->postings,
-                    data->cache[cache_key].diff_id,
-                    new_postings,
-                    GRN_OBJ_SET);
-
-  data->cache[cache_key].token_id = GRN_ID_NIL;
-  data->cache[cache_key].diff_id = GRN_ID_NIL;
-  data->cache[cache_key].offset = 0;
-  GRN_BULK_REWIND(&(data->cache[cache_key].postings));
-}
-
-static void
-grn_index_column_diff_cache_flush_all(grn_ctx *ctx,
-                                      grn_index_column_diff_data *data)
-{
-  for (size_t i = 0; i < data->cache_size; i++) {
-    grn_index_column_diff_cache_flush(ctx, data, i);
+  switch (compared) {
+  case GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL :
+    return '=';
+  case GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER :
+    return '>';
+  case GRN_INDEX_COLUMN_DIFF_COMPARED_LESS :
+    return '<';
+  default :
+    return '?';
   }
 }
 
-static grn_obj *
-grn_index_column_diff_cache_allocate(grn_ctx *ctx,
-                                     grn_index_column_diff_data *data)
-{
-  const grn_id token_id = data->current.token_id;
-  const size_t cache_key =
-    grn_index_column_diff_cache_compute_key(ctx, data, token_id);
-  grn_index_column_diff_cache_flush(ctx, data, cache_key);
-  data->cache[cache_key].token_id = token_id;
-  data->cache[cache_key].diff_id = data->current.diff_id;
-  return &(data->cache[cache_key].postings);
-}
-
-static const uint32_t *
-grn_index_column_diff_cache_get(grn_ctx *ctx,
-                                grn_index_column_diff_data *data,
-                                size_t *n_postings)
-{
-  const grn_id token_id = data->current.token_id;
-  const size_t cache_key
-    = grn_index_column_diff_cache_compute_key(ctx, data, token_id);
-  const grn_id cached_token_id = data->cache[cache_key].token_id;
-  grn_obj *postings;
-  if (cached_token_id == token_id) {
-    postings = &(data->cache[cache_key].postings);
-  } else {
-    postings = grn_index_column_diff_cache_allocate(ctx, data);
-    const grn_id diff_id = data->current.diff_id;
-    grn_obj_get_value(ctx, data->postings, diff_id, postings);
-  }
-  const size_t n_posting_elements = data->n_posting_elements;
-  const size_t offset = data->cache[cache_key].offset;
-  *n_postings =
-    (GRN_UINT32_VECTOR_SIZE(postings) / n_posting_elements) -
-    offset;
-  return ((uint32_t *)GRN_BULK_HEAD(postings)) + (n_posting_elements * offset);
-}
-
-static void
-grn_index_column_diff_cache_remove_posting(grn_ctx *ctx,
-                                           grn_index_column_diff_data *data,
-                                           size_t nth_posting)
-{
-  const grn_id token_id = data->current.token_id;
-  const size_t cache_key =
-    grn_index_column_diff_cache_compute_key(ctx, data, token_id);
-  const size_t n_posting_elements = data->n_posting_elements;
-  size_t n_postings = 0;
-  const uint32_t *postings =
-    grn_index_column_diff_cache_get(ctx, data, &n_postings);
-
-  if (nth_posting > 0) {
-    grn_obj *remains = &(data->buffers.postings);
-    GRN_BULK_REWIND(remains);
-    const size_t posting_size = sizeof(uint32_t) * n_posting_elements;
-    grn_bulk_write(ctx,
-                   remains,
-                   (const char *)postings,
-                   posting_size * nth_posting);
-    const grn_id diff_id = data->current.diff_id;
-    grn_obj_set_value(ctx, data->remains, diff_id, remains, GRN_OBJ_APPEND);
-  }
-  data->cache[cache_key].offset += nth_posting + 1;
-}
-
-static const uint32_t *
-grn_index_column_diff_get_postings(grn_ctx *ctx,
-                                   grn_index_column_diff_data *data,
-                                   size_t *n_postings)
-{
-  if (!data->current.is_new_diff) {
-    return grn_index_column_diff_cache_get(ctx, data, n_postings);
-  }
-
-  grn_obj *postings = grn_index_column_diff_cache_allocate(ctx, data);
-  const grn_id token_id = data->current.token_id;
-  const unsigned int ii_cursor_flags = 0;
-  grn_ii_cursor *ii_cursor = grn_ii_cursor_open(ctx,
-                                                data->ii,
-                                                token_id,
-                                                GRN_ID_NIL,
-                                                GRN_ID_MAX,
-                                                data->index.n_elements,
-                                                ii_cursor_flags);
-  if (ii_cursor) {
-    const grn_bool with_section = data->index.with_section;
-    const grn_bool with_position = data->index.with_position;
-    if (with_position) {
-      while (grn_ii_cursor_next(ctx, ii_cursor)) {
-        grn_posting *posting;
-        while ((posting = grn_ii_cursor_next_pos(ctx, ii_cursor))) {
-          GRN_UINT32_PUT(ctx, postings, posting->rid);
-          if (with_section) {
-            GRN_UINT32_PUT(ctx, postings, posting->sid);
-          }
-          GRN_UINT32_PUT(ctx, postings, posting->pos);
-        }
-      }
-    } else {
-      grn_posting *posting;
-      while ((posting = grn_ii_cursor_next(ctx, ii_cursor))) {
-        GRN_UINT32_PUT(ctx, postings, posting->rid);
-        if (with_section) {
-          GRN_UINT32_PUT(ctx, postings, posting->sid);
-        }
-      }
-    }
-    grn_ii_cursor_close(ctx, ii_cursor);
-  }
-  *n_postings = GRN_UINT32_VECTOR_SIZE(postings) / data->n_posting_elements;
-  return (const uint32_t *)(GRN_BULK_HEAD(postings));
-}
-
-static int
+static grn_index_column_diff_compared
 grn_index_column_diff_compare_posting(grn_ctx *ctx,
                                       grn_index_column_diff_data *data,
-                                      const uint32_t *posting)
+                                      const grn_posting *posting)
 {
   const grn_bool with_section = data->index.with_section;
   const grn_bool with_position = data->index.with_position;
   const grn_posting *current_posting = &(data->current.posting);
 
-  size_t i = 0;
-  const grn_id record_id = posting[i];
-  if (record_id < current_posting->rid) {
-    return -1;
-  } else if (record_id > current_posting->rid) {
-    return 1;
+  if (posting->rid < current_posting->rid) {
+    return GRN_INDEX_COLUMN_DIFF_COMPARED_LESS;
+  } else if (posting->rid > current_posting->rid) {
+    return GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER;
   }
 
   if (with_section) {
-    i++;
-    const uint32_t section_id = posting[i];
-    if (section_id < current_posting->sid) {
-      return -1;
-    } else if (section_id > current_posting->sid) {
-      return 1;
+    if (posting->sid < current_posting->sid) {
+      return GRN_INDEX_COLUMN_DIFF_COMPARED_LESS;
+    } else if (posting->sid > current_posting->sid) {
+      return GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER;
     }
   }
 
   if (with_position) {
-    i++;
-    const uint32_t position = posting[i];
-    if (position < current_posting->pos) {
-      return -1;
-    } else if (position > current_posting->pos) {
-      return 1;
+    if (posting->pos < current_posting->pos) {
+      return GRN_INDEX_COLUMN_DIFF_COMPARED_LESS;
+    } else if (posting->pos > current_posting->pos) {
+      return GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER;
     }
   }
 
-  return 0;
+  return GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL;
 }
 
-static int64_t
-grn_index_column_diff_find_posting(grn_ctx *ctx,
-                                   grn_index_column_diff_data *data)
+static int
+grn_index_column_diff_token_id_width(grn_ctx *ctx,
+                                     grn_index_column_diff_data *data)
 {
-  size_t n_postings = 0;
-  const uint32_t *postings =
-    grn_index_column_diff_get_postings(ctx, data, &n_postings);
-  const size_t n_posting_elements = data->n_posting_elements;
-  for (size_t i = 0; i < n_postings; i++) {
-    const uint32_t *posting = postings + (n_posting_elements * i);
-    const int compared =
-      grn_index_column_diff_compare_posting(ctx, data, posting);
-    if (compared == 0) {
-      return i;
-    } else if (compared > 0) {
-      return -1;
+  unsigned int n_tokens = grn_table_size(ctx, data->lexicon);
+  if (n_tokens == 0) {
+    return 0;
+  } else {
+    return floor(log10(n_tokens)) + 1;
+  }
+}
+
+static void
+grn_index_column_diff_posting_list_init(grn_ctx *ctx,
+                                        grn_index_column_diff_data *data,
+                                        grn_index_column_diff_posting_list *posting_list)
+{
+  posting_list->token_id = data->current.token_id;
+  GRN_UINT32_INIT(&(posting_list->remains), GRN_OBJ_VECTOR);
+  GRN_UINT32_INIT(&(posting_list->missings), GRN_OBJ_VECTOR);
+  GRN_UINT32_INIT(&(posting_list->look_ahead), GRN_OBJ_VECTOR);
+  posting_list->look_ahead_offset = 0;
+  posting_list->last_posting.tf = 0;
+
+  const unsigned int ii_cursor_flags = 0;
+  posting_list->cursor = grn_ii_cursor_open(ctx,
+                                            data->ii,
+                                            posting_list->token_id,
+                                            GRN_ID_NIL,
+                                            GRN_ID_MAX,
+                                            data->index.n_elements,
+                                            ii_cursor_flags);
+  posting_list->need_cursor_next = true;
+  for (size_t i = 0; i < N_GRN_INDEX_COLUMN_DIFF_PERSISTENT_TYPES; i++) {
+    posting_list->n_postings[i] = 0;
+    posting_list->thresholds[i] = 4096;
+  }
+}
+
+static void
+grn_index_column_diff_posting_list_fin(grn_ctx *ctx,
+                                       grn_index_column_diff_data *data,
+                                       grn_index_column_diff_posting_list *posting_list)
+{
+  grn_obj *remains = &(posting_list->remains);
+  grn_obj *missings = &(posting_list->missings);
+  {
+    grn_obj *look_ahead = &(posting_list->look_ahead);
+    const size_t offset = posting_list->look_ahead_offset * sizeof(uint32_t);
+    grn_bulk_write(ctx,
+                   remains,
+                   GRN_BULK_HEAD(look_ahead) + offset,
+                   GRN_BULK_VSIZE(look_ahead) - offset);
+    GRN_OBJ_FIN(ctx, look_ahead);
+  }
+
+  grn_ii_cursor *cursor = posting_list->cursor;
+  if (cursor) {
+    const grn_bool with_section = data->index.with_section;
+    const grn_bool with_position = data->index.with_position;
+    if (with_position) {
+      while (!posting_list->need_cursor_next ||
+             grn_ii_cursor_next(ctx, cursor)) {
+        grn_posting *posting;
+        while ((posting = grn_ii_cursor_next_pos(ctx, cursor))) {
+          posting_list->need_cursor_next = false;
+          GRN_UINT32_PUT(ctx, remains, posting->rid);
+          if (with_section) {
+            GRN_UINT32_PUT(ctx, remains, posting->sid);
+          }
+          GRN_UINT32_PUT(ctx, remains, posting->pos);
+        }
+        posting_list->need_cursor_next = true;
+      }
+    } else {
+      grn_posting *posting;
+      while ((posting = grn_ii_cursor_next(ctx, cursor))) {
+        GRN_UINT32_PUT(ctx, remains, posting->rid);
+        if (with_section) {
+          GRN_UINT32_PUT(ctx, remains, posting->sid);
+        }
+      }
+    }
+    grn_ii_cursor_close(ctx, cursor);
+    posting_list->cursor = NULL;
+  }
+
+  if (GRN_BULK_VSIZE(remains) > 0 || GRN_BULK_VSIZE(missings) > 0) {
+    const grn_id diff_id = grn_table_add(ctx,
+                                         data->diff,
+                                         &(posting_list->token_id),
+                                         sizeof(grn_id),
+                                         NULL);
+    if (diff_id != GRN_ID_NIL) {
+      if (GRN_BULK_VSIZE(remains) > 0) {
+        grn_obj_set_value(ctx,
+                          data->remains,
+                          diff_id,
+                          remains,
+                          GRN_OBJ_APPEND);
+      }
+      if (GRN_BULK_VSIZE(missings) > 0) {
+        grn_obj_set_value(ctx,
+                          data->missings,
+                          diff_id,
+                          missings,
+                          GRN_OBJ_APPEND);
+      }
     }
   }
-  return -1;
+  GRN_OBJ_FIN(ctx, remains);
+  GRN_OBJ_FIN(ctx, missings);
+}
+
+static void
+grn_index_column_diff_posting_list_put(grn_ctx *ctx,
+                                       grn_index_column_diff_data *data,
+                                       grn_index_column_diff_posting_list *posting_list,
+                                       grn_index_column_diff_type type,
+                                       const grn_posting *posting)
+{
+  grn_obj *column;
+  grn_obj *postings;
+  switch (type) {
+  case GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS :
+    column = data->remains;
+    postings = &(posting_list->remains);
+    break;
+  case GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS :
+    column = data->missings;
+    postings = &(posting_list->missings);
+    break;
+  default :
+    column = NULL;
+    postings = &(posting_list->look_ahead);
+    break;
+  }
+
+  GRN_UINT32_PUT(ctx, postings, posting->rid);
+  if (data->index.with_section) {
+    GRN_UINT32_PUT(ctx, postings, posting->sid);
+  }
+  if (data->index.with_position) {
+    GRN_UINT32_PUT(ctx, postings, posting->pos);
+  }
+
+  if (!column) {
+    return;
+  }
+
+  posting_list->n_postings[type]++;
+  const size_t current_n_postings =
+    GRN_UINT32_VECTOR_SIZE(postings) / data->n_posting_elements;
+  if (current_n_postings < posting_list->thresholds[type]) {
+    return;
+  }
+
+  const grn_id diff_id = grn_table_add(ctx,
+                                       data->diff,
+                                       &(posting_list->token_id),
+                                       sizeof(grn_id),
+                                       NULL);
+  if (diff_id == GRN_ID_NIL) {
+    return;
+  }
+  grn_obj_set_value(ctx, column, diff_id, postings, GRN_OBJ_APPEND);
+  if (ctx->rc == GRN_SUCCESS) {
+    GRN_BULK_REWIND(postings);
+  }
+  posting_list->thresholds[type] *= 2;
+  GRN_LOG(ctx,
+          GRN_LOG_DUMP,
+          "[index-column][diff][append][%.*s][%u][%*s] "
+          "%" GRN_FMT_SIZE "(+%" GRN_FMT_SIZE ")",
+          data->index.name_size,
+          data->index.name,
+          grn_index_column_diff_token_id_width(ctx, data),
+          data->current.token_id,
+          grn_index_column_diff_type_to_string(type),
+          posting_list->n_postings[type],
+          current_n_postings);
+}
+
+static void
+grn_index_column_diff_posting_list_consume_look_ahead(
+  grn_ctx *ctx,
+  grn_index_column_diff_data *data,
+  grn_index_column_diff_posting_list *posting_list)
+{
+  grn_obj *look_ahead = &(posting_list->look_ahead);
+  posting_list->look_ahead_offset++;
+  if (posting_list->look_ahead_offset ==
+      (GRN_UINT32_VECTOR_SIZE(look_ahead) / data->n_posting_elements)) {
+    GRN_BULK_REWIND(look_ahead);
+    posting_list->look_ahead_offset = 0;
+  }
 }
 
 static void
 grn_index_column_diff_process_token_id(grn_ctx *ctx,
                                        grn_index_column_diff_data *data)
 {
-  {
-    int added = 0;
-    data->current.diff_id =
-      grn_table_add(ctx,
-                    data->diff,
-                    &(data->current.token_id), sizeof(grn_id),
-                    &added);
-    data->current.is_new_diff = (added != 0);
+  void *value = NULL;
+  int added = 0;
+  const grn_id posting_list_id =
+    grn_hash_add(ctx,
+                 data->posting_lists,
+                 &(data->current.token_id), sizeof(grn_id),
+                 &value,
+                 &added);
+  if (posting_list_id == GRN_ID_NIL) {
+    grn_rc rc = ctx->rc;
+    if (rc == GRN_SUCCESS) {
+      rc = GRN_UNKNOWN_ERROR;
+    }
+    char message[GRN_CTX_MSGSIZE];
+    grn_strcpy(message, GRN_CTX_MSGSIZE, ctx->errbuf);
+    ERR(rc,
+        "[index-column][diff] failed to add a record to keep a posting list: "
+        "%u: %s",
+        data->current.token_id,
+        message);
+    return;
   }
 
-  const int64_t nth_posting =
-    grn_index_column_diff_find_posting(ctx, data);
-  if (nth_posting >= 0) {
-    grn_index_column_diff_cache_remove_posting(ctx,
-                                               data,
-                                               (size_t)nth_posting);
-  } else {
-    grn_obj *postings = &(data->buffers.postings);
-    GRN_BULK_REWIND(postings);
-    const grn_posting *posting = &(data->current.posting);
-    GRN_UINT32_PUT(ctx, postings, posting->rid);
-    if (data->index.with_section) {
-      GRN_UINT32_PUT(ctx, postings, posting->sid);
-    }
-    if (data->index.with_position) {
-      GRN_UINT32_PUT(ctx, postings, posting->pos);
-    }
-    grn_obj_set_value(ctx,
-                      data->missings,
-                      data->current.diff_id,
-                      postings,
-                      GRN_OBJ_APPEND);
+  grn_index_column_diff_posting_list *posting_list = value;
+  if (added) {
+    grn_index_column_diff_posting_list_init(ctx, data, posting_list);
   }
+
+  const grn_bool with_section = data->index.with_section;
+  const grn_bool with_position = data->index.with_position;
+
+  if (posting_list->last_posting.tf > 0) {
+    const grn_index_column_diff_compared compared =
+      grn_index_column_diff_compare_posting(ctx,
+                                            data,
+                                            &(posting_list->last_posting));
+    GRN_LOG(ctx,
+            GRN_LOG_DUMP,
+            "[index-column][diff][process][%.*s][%*u] "
+            "    last[%u:%u:%u:%u] %c current[%u:%u:%u]",
+            data->index.name_size,
+            data->index.name,
+            grn_index_column_diff_token_id_width(ctx, data),
+            data->current.token_id,
+            posting_list->last_posting.rid,
+            posting_list->last_posting.sid,
+            posting_list->last_posting.pos,
+            posting_list->last_posting.tf,
+            grn_index_column_diff_compared_to_char(compared),
+            data->current.posting.rid,
+            data->current.posting.sid,
+            data->current.posting.pos);
+    if (compared == GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL) {
+      posting_list->last_posting.tf--;
+      return;
+    } else {
+      posting_list->last_posting.tf = 0;
+    }
+  }
+
+  grn_obj *look_ahead = &(posting_list->look_ahead);
+  const size_t look_ahead_offset = posting_list->look_ahead_offset;
+
+  const size_t n_posting_elements = data->n_posting_elements;
+  const size_t n_postings =
+    GRN_UINT32_VECTOR_SIZE(look_ahead) /
+    n_posting_elements;
+  const uint32_t *raw_postings = (const uint32_t *)GRN_BULK_HEAD(look_ahead);
+  for (size_t i = look_ahead_offset; i < n_postings; i++) {
+    const size_t offset = i * data->n_posting_elements;
+    const uint32_t *raw_posting = raw_postings + offset;
+    grn_posting posting;
+    size_t i = 0;
+    posting.rid = raw_posting[i++];
+    if (with_section) {
+      posting.sid = raw_posting[i++];
+    }
+    if (with_position) {
+      posting.pos = raw_posting[i++];
+    }
+    const grn_index_column_diff_compared compared =
+      grn_index_column_diff_compare_posting(ctx, data, &posting);
+    GRN_LOG(ctx,
+            GRN_LOG_DUMP,
+            "[index-column][diff][process][%.*s][%*u] "
+            "look-ahead[%u:%u:%u] %c current[%u:%u:%u]",
+            data->index.name_size,
+            data->index.name,
+            grn_index_column_diff_token_id_width(ctx, data),
+            data->current.token_id,
+            posting.rid,
+            posting.sid,
+            posting.pos,
+            grn_index_column_diff_compared_to_char(compared),
+            data->current.posting.rid,
+            data->current.posting.sid,
+            data->current.posting.pos);
+    switch (compared) {
+    case GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL :
+      grn_index_column_diff_posting_list_consume_look_ahead(ctx,
+                                                            data,
+                                                            posting_list);
+      return;
+    case GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER :
+      grn_index_column_diff_posting_list_put(ctx,
+                                             data,
+                                             posting_list,
+                                             GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS,
+                                             &(data->current.posting));
+      return;
+    default :
+      grn_index_column_diff_posting_list_put(ctx,
+                                             data,
+                                             posting_list,
+                                             GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS,
+                                             &posting);
+      grn_index_column_diff_posting_list_consume_look_ahead(ctx,
+                                                            data,
+                                                            posting_list);
+      break;
+    }
+  }
+
+  grn_ii_cursor *cursor = posting_list->cursor;
+  if (cursor) {
+    if (with_position) {
+      while (!posting_list->need_cursor_next ||
+             grn_ii_cursor_next(ctx, cursor)) {
+        grn_posting *posting;
+        while ((posting = grn_ii_cursor_next_pos(ctx, cursor))) {
+          posting_list->need_cursor_next = false;
+          const grn_index_column_diff_compared compared =
+            grn_index_column_diff_compare_posting(ctx, data, posting);
+          GRN_LOG(ctx,
+                  GRN_LOG_DUMP,
+                  "[index-column][diff][process][%.*s][%*u] "
+                  "      read[%u:%u:%u] %c current[%u:%u:%u]",
+                  data->index.name_size,
+                  data->index.name,
+                  grn_index_column_diff_token_id_width(ctx, data),
+                  data->current.token_id,
+                  posting->rid,
+                  posting->sid,
+                  posting->pos,
+                  grn_index_column_diff_compared_to_char(compared),
+                  data->current.posting.rid,
+                  data->current.posting.sid,
+                  data->current.posting.pos);
+          switch (compared) {
+          case GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL :
+            return;
+          case GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER :
+            grn_index_column_diff_posting_list_put(
+              ctx,
+              data,
+              posting_list,
+              GRN_INDEX_COLUMN_DIFF_TYPE_LOOK_AHEAD,
+              posting);
+            grn_index_column_diff_posting_list_put(
+              ctx,
+              data,
+              posting_list,
+              GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS,
+              &(data->current.posting));
+            return;
+          default :
+            grn_index_column_diff_posting_list_put(
+              ctx,
+              data,
+              posting_list,
+              GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS,
+              posting);
+            break;
+          }
+        }
+        posting_list->need_cursor_next = true;
+      }
+    } else {
+      grn_posting *posting;
+      while ((posting = grn_ii_cursor_next(ctx, cursor))) {
+        const grn_index_column_diff_compared compared =
+          grn_index_column_diff_compare_posting(ctx, data, posting);
+        GRN_LOG(ctx,
+                GRN_LOG_DUMP,
+                "[index-column][diff][process][%.*s][%*u] "
+                "    read[%u:%u:%u:%u] %c current[%u:%u:%u]",
+                data->index.name_size,
+                data->index.name,
+                grn_index_column_diff_token_id_width(ctx, data),
+                data->current.token_id,
+                posting->rid,
+                posting->sid,
+                posting->pos,
+                posting->tf,
+                grn_index_column_diff_compared_to_char(compared),
+                data->current.posting.rid,
+                data->current.posting.sid,
+                data->current.posting.pos);
+        switch (compared) {
+        case GRN_INDEX_COLUMN_DIFF_COMPARED_EQUAL :
+          if (!with_position && posting->tf > 0) {
+            posting_list->last_posting = *posting;
+            posting_list->last_posting.tf--;
+          }
+          return;
+        case GRN_INDEX_COLUMN_DIFF_COMPARED_GREATER :
+          grn_index_column_diff_posting_list_put(
+            ctx,
+            data,
+            posting_list,
+            GRN_INDEX_COLUMN_DIFF_TYPE_LOOK_AHEAD,
+            posting);
+          grn_index_column_diff_posting_list_put(
+            ctx,
+            data,
+            posting_list,
+            GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS,
+            &(data->current.posting));
+          return;
+        default :
+          grn_index_column_diff_posting_list_put(
+            ctx,
+            data,
+            posting_list,
+            GRN_INDEX_COLUMN_DIFF_TYPE_REMAINS,
+            posting);
+          break;
+        }
+      }
+    }
+  }
+
+  GRN_LOG(ctx,
+          GRN_LOG_DUMP,
+          "[index-column][diff][process][%.*s][%*u] "
+          " no posting list:    current[%u:%u:%u]",
+          data->index.name_size,
+          data->index.name,
+          grn_index_column_diff_token_id_width(ctx, data),
+          data->current.token_id,
+          data->current.posting.rid,
+          data->current.posting.sid,
+          data->current.posting.pos);
+  grn_index_column_diff_posting_list_put(ctx,
+                                         data,
+                                         posting_list,
+                                         GRN_INDEX_COLUMN_DIFF_TYPE_MISSINGS,
+                                         &(data->current.posting));
 }
 
 static void
@@ -800,6 +1007,9 @@ grn_index_column_diff_process_token(grn_ctx *ctx,
     }
 
     grn_index_column_diff_process_token_id(ctx, data);
+    if (ctx->rc != GRN_SUCCESS) {
+      break;
+    }
   }
   grn_token_cursor_close(ctx, token_cursor);
 }
@@ -811,7 +1021,6 @@ grn_index_column_diff_compute(grn_ctx *ctx,
   grn_obj *source_columns = &(data->source_columns);
   const size_t n_source_columns = GRN_PTR_VECTOR_SIZE(source_columns);
   grn_obj *value = &(data->buffers.value);
-  grn_obj *postings = &(data->buffers.postings);
 
   grn_index_column_diff_init_progress(ctx, data);
 
@@ -830,6 +1039,9 @@ grn_index_column_diff_compute(grn_ctx *ctx,
 
       GRN_BULK_REWIND(value);
       grn_obj_get_value(ctx, source, id, value);
+      if (ctx->rc != GRN_SUCCESS) {
+        break;
+      }
 
       switch (value->header.type) {
       case GRN_VECTOR :
@@ -844,6 +1056,9 @@ grn_index_column_diff_compute(grn_ctx *ctx,
                                                 data,
                                                 element,
                                                 element_size);
+            if (ctx->rc != GRN_SUCCESS) {
+              break;
+            }
           }
         }
         break;
@@ -853,9 +1068,15 @@ grn_index_column_diff_compute(grn_ctx *ctx,
           for (size_t j = 0; j < n_elements; j++) {
             const grn_id element =
               grn_uvector_get_element(ctx, value, j, NULL);
+            if (element == GRN_ID_NIL) {
+              continue;
+            }
             data->current.token_id = element;
             data->current.posting.pos = 0;
             grn_index_column_diff_process_token_id(ctx, data);
+            if (ctx->rc != GRN_SUCCESS) {
+              break;
+            }
           }
         } else {
           const size_t n_elements = grn_uvector_size(ctx, value);
@@ -869,6 +1090,9 @@ grn_index_column_diff_compute(grn_ctx *ctx,
             }
             data->current.posting.pos = 0;
             grn_index_column_diff_process_token_id(ctx, data);
+            if (ctx->rc != GRN_SUCCESS) {
+              break;
+            }
           }
         }
         break;
@@ -876,8 +1100,10 @@ grn_index_column_diff_compute(grn_ctx *ctx,
         if (GRN_BULK_VSIZE(value) > 0) {
           if (is_reference) {
             data->current.token_id = GRN_RECORD_VALUE(value);
-            data->current.posting.pos = 0;
-            grn_index_column_diff_process_token_id(ctx, data);
+            if (data->current.token_id != GRN_ID_NIL) {
+              data->current.posting.pos = 0;
+              grn_index_column_diff_process_token_id(ctx, data);
+            }
           } else {
             grn_index_column_diff_process_token(ctx,
                                                 data,
@@ -889,34 +1115,23 @@ grn_index_column_diff_compute(grn_ctx *ctx,
       default :
         break;
       }
+
+      if (ctx->rc != GRN_SUCCESS) {
+        break;
+      }
+    }
+    if (ctx->rc != GRN_SUCCESS) {
+      break;
     }
   } GRN_TABLE_EACH_END(ctx, cursor);
 
-  grn_index_column_diff_cache_flush_all(ctx, data);
-
-  {
-    GRN_TABLE_EACH_BEGIN(ctx, data->diff, cursor, id) {
-      GRN_BULK_REWIND(postings);
-      grn_obj_get_value(ctx, data->postings, id, postings);
-      if (GRN_UINT32_VECTOR_SIZE(postings) > 0) {
-        grn_obj_set_value(ctx, data->remains, id, postings, GRN_OBJ_APPEND);
-        continue;
-      }
-      GRN_BULK_REWIND(postings);
-      grn_obj_get_value(ctx, data->remains, id, postings);
-      if (GRN_UINT32_VECTOR_SIZE(postings) > 0) {
-        continue;
-      }
-      GRN_BULK_REWIND(postings);
-      grn_obj_get_value(ctx, data->missings, id, postings);
-      if (GRN_UINT32_VECTOR_SIZE(postings) > 0) {
-        continue;
-      }
-      grn_table_cursor_delete(ctx, cursor);
-    } GRN_TABLE_EACH_END(ctx, cursor);
-  }
-
-  grn_obj_close(ctx, data->postings);
+  GRN_HASH_EACH_BEGIN(ctx, data->posting_lists, cursor, id) {
+    void *value;
+    grn_hash_cursor_get_value(ctx, cursor, &value);
+    grn_index_column_diff_posting_list *posting_list = value;
+    grn_index_column_diff_posting_list_fin(ctx, data, posting_list);
+  } GRN_HASH_EACH_END(ctx, cursor);
+  grn_hash_close(ctx, data->posting_lists);
 }
 
 grn_rc
@@ -989,7 +1204,6 @@ grn_index_column_diff(grn_ctx *ctx,
   }
 
   data.lexicon = grn_ctx_at(ctx, index_column->header.domain);
-  grn_index_column_diff_data_init_cache(ctx, &data);
   if (ctx->rc != GRN_SUCCESS) {
     goto exit;
   }
@@ -1010,23 +1224,24 @@ grn_index_column_diff(grn_ctx *ctx,
         message);
     goto exit;
   }
-  data.postings = grn_column_create(ctx,
-                                    data.diff,
-                                    postings_column_name,
-                                    strlen(postings_column_name),
-                                    NULL,
-                                    GRN_OBJ_COLUMN_VECTOR,
-                                    grn_ctx_at(ctx, GRN_DB_UINT32));
-  if (!data.postings) {
+  data.posting_lists =
+    grn_hash_create(ctx,
+                    NULL,
+                    sizeof(grn_id),
+                    sizeof(grn_index_column_diff_posting_list),
+                    GRN_OBJ_TABLE_HASH_KEY);
+  if (!data.posting_lists) {
     char message[GRN_CTX_MSGSIZE];
     grn_strcpy(message, GRN_CTX_MSGSIZE, ctx->errbuf);
     ERR(GRN_INVALID_ARGUMENT,
-        "[index-column][diff] failed to create postings column: <%.*s>: %s",
+        "[index-column][diff] failed to create hash table for posting lists: "
+        "<%.*s>: %s",
         data.index.name_size,
         data.index.name,
         message);
     goto exit;
   }
+  DB_OBJ(data.posting_lists)->header.domain = grn_obj_id(ctx, data.diff);
   data.remains = grn_column_create(ctx,
                                    data.diff,
                                    remains_column_name,
