@@ -65,30 +65,77 @@ if ! type bundle 2>&1 > /dev/null; then
   $RUBY -S gem install bundler
 fi
 
-grntest_dir="$BUILD_DIR/grntest"
+grntest_dir="$SOURCE_DIR/grntest"
 if ! test -d "$grntest_dir"; then
+  grntest_dir="$BUILD_DIR/grntest"
   git clone --depth 1 git://github.com/groonga/grntest.git "$grntest_dir"
   (cd "$grntest_dir" && bundle install)
-else
-  (cd "$grntest_dir";
-   if [ "Gemfile" -nt "Gemfile.lock" ]; then
-     $RUBY -S bundle update
-   fi)
 fi
+(cd "$grntest_dir";
+ if [ "Gemfile" -nt "Gemfile.lock" ]; then
+   $RUBY -S bundle update
+ fi)
 
-groonga_command_dir="$BUILD_DIR/groonga-command"
+groonga_command_dir="$SOURCE_DIR/groonga-command"
+if ! test -d "$groonga_command_dir"; then
+  groonga_command_dir="$BUILD_DIR/groonga-command"
+fi
 if ! test -d "$groonga_command_dir"; then
   git clone --depth 1 \
-      git://github.com/groonga/groonga-command "$groonga_command_dir"
+      https://github.com/groonga/groonga-command.git \
+      "$groonga_command_dir"
 fi
 
-groonga_command_parser_dir="$BUILD_DIR/groonga-command-parser"
+groonga_command_parser_dir="$SOURCE_DIR/groonga-command-parser"
+if ! test -d "$groonga_command_parser_dir"; then
+  groonga_command_parser_dir="$BUILD_DIR/groonga-command-parser"
+fi
 if ! test -d "$groonga_command_parser_dir"; then
   git clone --depth 1 \
-      git://github.com/groonga/groonga-command-parser \
+      https://github.com/groonga/groonga-command-parser.git \
       "$groonga_command_parser_dir"
 fi
 
+gqtp_dir="$SOURCE_DIR/gqtp"
+if ! test -d "$gqtp_dir"; then
+  gqtp_dir="$BUILD_DIR/gqtp"
+fi
+if ! test -d "$gqtp_dir"; then
+  git clone --depth 1 \
+      https://github.com/ranguba/gqtp.git \
+      "$gqtp_dir"
+fi
+
+groonga_client_dir="$SOURCE_DIR/groonga-client"
+if ! test -d "$groonga_client_dir"; then
+  groonga_client_dir="$BUILD_DIR/groonga-client"
+fi
+if ! test -d "$groonga_client_dir"; then
+  git clone --depth 1 \
+      https://github.com/ranguba/groonga-client.git \
+      "$groonga_client_dir"
+fi
+
+groonga_log_dir="$SOURCE_DIR/groonga-log"
+if ! test -d "$groonga_log_dir"; then
+  groonga_log_dir="$BUILD_DIR/groonga-log"
+fi
+if ! test -d "$groonga_log_dir"; then
+  git clone --depth 1 \
+      https://github.com/groonga/groonga-log.git \
+      "$groonga_log_dir"
+fi
+
+groonga_query_log_dir="$SOURCE_DIR/groonga-query-log"
+if ! test -d "$groonga_query_log_dir"; then
+  groonga_query_log_dir="$BUILD_DIR/groonga-query-log"
+fi
+if ! test -d "$groonga_query_log_dir"; then
+  git clone --depth 1 \
+      https://github.com/groonga/groonga-query-log.git \
+      "$groonga_query_log_dir"
+fi
+
 have_targets="false"
 use_gdb="false"
 use_valgrind="false"
@@ -150,6 +197,10 @@ $RUBY \
   -I "$grntest_dir/lib" \
   -I "$groonga_command_dir/lib" \
   -I "$groonga_command_parser_dir/lib" \
+  -I "$gqtp_dir/lib" \
+  -I "$groonga_client_dir/lib" \
+  -I "$groonga_log_dir/lib" \
+  -I "$groonga_query_log_dir/lib" \
   "$grntest_dir/bin/grntest" \
   --groonga "$GROONGA" \
   --groonga-httpd "$GROONGA_HTTPD" \
